@@ -368,9 +368,11 @@ function(e1, e2)
             e2 <- e3
         }
         ## Find duplicated indices.
-        pos <- match(paste(e2$i, e2$j, sep = "\r"),
-                     paste(e1$i, e1$j, sep = "\r"),
-                     nomatch = 0L)
+        ## pos <- match(paste(e2$i, e2$j, sep = "\r"),
+        ##              paste(e1$i, e1$j, sep = "\r"),
+        ##              nomatch = 0L)
+        pos <- .Call(R_match_matrix, cbind(e1$i, e1$j),
+                                     cbind(e2$i, e2$j), 0L)[[2L]]
         ind <- which(pos > 0L)
         if(!all(is.finite(e1$v)) || !all(is.finite(e2$v))) {
 	    ## Augment and reduce
@@ -414,9 +416,11 @@ function(e1, e2)
         e2 <- e3
     }
     ## Find duplicated indices.
-    pos <- match(paste(e2$i, e2$j, sep = "\r"),
-                 paste(e1$i, e1$j, sep = "\r"),
-                 nomatch = 0L)
+    ## pos <- match(paste(e2$i, e2$j, sep = "\r"),
+    ##              paste(e1$i, e1$j, sep = "\r"),
+    ##              nomatch = 0L)
+    pos <- .Call(R_match_matrix, cbind(e1$i, e1$j),
+                                 cbind(e2$i, e2$j), 0L)[[2L]]
     ind <- which(pos == 0L)
     ## Notice 0 + special value = special value.
     e1$v[pos] <- e1$v[pos] + e2$v[pos > 0L]
@@ -505,7 +509,7 @@ function(x, value)
     }
     ## See the constructor (above).
     if(is.null(value))
-        x["dimnames"] <- list(NULL)
+        x$dimnames <- NULL
     else
         x$dimnames <- dnx
     x
@@ -701,8 +705,20 @@ function(x, i, j, drop = FALSE)
             pj[j] <- seq_len(nc)
         }
 
-        if(!is.null(dnx <- x$dimnames))
-            dnx[] <- list(dnx[[1L]][i], dnx[[2L]][j])
+        if(!is.null(dnx <- x$dimnames)) {
+	    if (!missing(i)) {
+		dnx[1L] <- list(dnx[[1L]][i])
+		if (!length(dnx[[1L]]))
+		    dnx[1L] <- list(NULL)
+	    }
+	    if (!missing(j)) {
+		dnx[2L] <- list(dnx[[2L]][j])
+		if (!length(dnx[[2L]]))
+		    dnx[2L] <- list(NULL)
+	    }
+	    if (!length(dnx[[1L]]) && !length(dnx[[2L]]))
+		dnx <- NULL
+	}
 
         i <- if(missing(i)) x$i[pos] else pi[x$i[pos]]
         j <- if(missing(j)) x$j[pos] else pj[x$j[pos]]
